#include <cmath>

#include "npstat/nm/MathUtils.hh"
#include "npstat/nm/SpecialFunctions.hh"
#include "npstat/nm/findRootUsingBisections.hh"
#include "npstat/nm/GaussHermiteQuadrature.hh"

#include "npstat/stat/EdgeworthSeries1D.hh"
#include "npstat/stat/distributionReadError.hh"

#include "geners/binaryIO.hh"
#include "geners/vectorIO.hh"

#define MAXCOEFFS 12
#define LSQR2 1.41421356237309504880169L
#define SQRTWOPIL 2.50662827463100050241577L

static long double ldgexceedance(const long double x)
{
    return erfcl(x/LSQR2)/2.0L;
}

static long double ldgcdf(const long double x)
{
    return erfcl(-x/LSQR2)/2.0L;
}

namespace npstat {
    EdgeworthSeries1D::EdgeworthSeries1D(
        const std::vector<double>& cumulants,
        const EdgeworthSeriesMethod m, const unsigned i_order,
        const bool i_slrMode)
        : cumulants_(cumulants), m_(m), order_(i_order), slrMode_(i_slrMode)
    {
        if (order_ > 4) throw std::invalid_argument(
            "In npstat::EdgeworthSeries1D constructor: "
            "order parameter out of range");
        if (order_)
        {
            const unsigned nMin = minNCumulants();
            if (cumulants_.size() < nMin)
                throw std::invalid_argument(
                    "In npstat::EdgeworthSeries1D constructor: "
                    "not enough cumulants provided");
            if (nMin > 1U)
                if (cumulants_[1] <= 0.0)
                    throw std::invalid_argument(
                        "In npstat::EdgeworthSeries1D constructor: "
                        "variance must be positive");
        }
    }

    double EdgeworthSeries1D::edgMean() const
    {
        if (minNCumulants() && m_ == EDGEWORTH_CLASSICAL)
            return cumulants_.at(0);
        else
            return 0.0;
    }

    double EdgeworthSeries1D::edgStdev() const
    {
        if (minNCumulants() > 1)
            return normalizedSigma();
        else
            return 1.0;
    }

    unsigned EdgeworthSeries1D::minNCumulants() const
    {
        if (slrMode_)
            return order_;
        else
            return order_ ? order_ + 2U : 0U;
    }

    bool EdgeworthSeries1D::isEqual(const AbsDistribution1D& o) const
    {
        const EdgeworthSeries1D& r = static_cast<const EdgeworthSeries1D&>(o);
        return cumulants_ == r.cumulants_ && m_ == r.m_ &&
               order_ == r.order_ && slrMode_ == r.slrMode_;
    }

    bool EdgeworthSeries1D::write(std::ostream& os) const
    {
        const unsigned m = static_cast<unsigned>(m_);
        const unsigned char slr = slrMode_;
        gs::write_pod(os, m);
        gs::write_pod(os, order_);
        gs::write_pod(os, slr);
        return gs::write_item(os, cumulants_, false) && !os.fail();
    }

    EdgeworthSeries1D* EdgeworthSeries1D::read(
        const gs::ClassId& id, std::istream& in)
    {
        static const gs::ClassId myClassId(
            gs::ClassId::makeId<EdgeworthSeries1D>());
        myClassId.ensureSameId(id);

        unsigned m, order;
        unsigned char slr;
        gs::read_pod(in, &m);
        gs::read_pod(in, &order);
        gs::read_pod(in, &slr);

        std::vector<double> cumulants;
        gs::restore_item(in, &cumulants, false);

        if (in.fail())
        {
            distributionReadError(in, classname());
            return 0;
        }
        else
            return new EdgeworthSeries1D(
                cumulants, static_cast<EdgeworthSeriesMethod>(m), order, slr);
    }

    double EdgeworthSeries1D::densityFactor(const double xin) const
    {
        long double coeffs[MAXCOEFFS+1];
        unsigned maxdeg = 0;

        if (order_)
        {
            if (slrMode_)
                hermiteCoeffsSLR(coeffs+1, &maxdeg);
            else
                hermiteCoeffsNormal(coeffs+1, &maxdeg);
            ++maxdeg;
        }
        coeffs[0] = 0.0L;
        return 1.0 + hermiteSeriesSumProb(coeffs, maxdeg, normalizedCoord(xin))/normalizedSigma();
    }

    double EdgeworthSeries1D::normalizedSigma() const
    {
        double sigma = 1.0;
        if (order_ > 1U && m_ == EDGEWORTH_CLASSICAL)
            sigma = sqrt(cumulants_.at(1));
        return sigma;
    }

    double EdgeworthSeries1D::density(const double xin) const
    {
        const double x = normalizedCoord(xin);
        return exp(-x*x/2)/SQRTWOPIL/normalizedSigma()*densityFactor(xin);
    }

    double EdgeworthSeries1D::cdf(const double xin) const
    {
        const double z = normalizedCoord(xin);
        return ldgcdf(z) - expl(-z*z/2.0)/SQRTWOPIL/normalizedSigma()*cdfFactor(xin);
    }

    double EdgeworthSeries1D::exceedance(const double xin) const
    {
        const double z = normalizedCoord(xin);
        return ldgexceedance(z) + expl(-z*z/2.0)/SQRTWOPIL/normalizedSigma()*cdfFactor(xin);
    }

    double EdgeworthSeries1D::quantile(const double r1) const
    {
        if (!(r1 >= 0.0 && r1 <= 1.0)) throw std::domain_error(
            "In npstat::EdgeworthSeries1D::quantile: "
            "cdf argument outside of [0, 1] interval");
        if (r1 == 0.0 || r1 == 1.0)
            return inverseGaussCdf(r1);
        const double x0 = inverseGaussCdf(0.0);
        const double x1 = inverseGaussCdf(1.0);
        CdfFunctor1D fcn(*this);
        double q = 0.0;
        if (!findRootUsingBisections(fcn, r1, x0, x1, 1.0e-15, &q))
            throw std::runtime_error("In npstat::EdgeworthSeries1D::quantile: "
                                     "root finding failed");
        return q;
    }

    double EdgeworthSeries1D::inverseExceedance(const double r1) const
    {
        if (!(r1 >= 0.0 && r1 <= 1.0)) throw std::domain_error(
            "In npstat::EdgeworthSeries1D::inverseExceedance: "
            "exceedance argument outside of [0, 1] interval");
        if (r1 == 0.0 || r1 == 1.0)
            return inverseGaussCdf(1.0 - r1);
        const double x0 = inverseGaussCdf(0.0);
        const double x1 = inverseGaussCdf(1.0);
        ExceedanceFunctor1D fcn(*this);
        double q = 0.0;
        if (!findRootUsingBisections(fcn, r1, x0, x1, 1.0e-15, &q))
            throw std::runtime_error("In npstat::EdgeworthSeries1D::inverseExceedance: "
                                     "root finding failed");
        return q;
    }

    double EdgeworthSeries1D::normalizedCoord(const double xin) const
    {
        double x = xin;
        if (order_)
        {
            switch (m_)
            {
            case EDGEWORTH_SEVERINI:
                break;

            case EDGEWORTH_CLASSICAL:
                {
                    double sigma = 1.0;
                    if (order_ > 1U)
                        sigma = sqrt(cumulants_.at(1));
                    x = (xin - cumulants_.at(0))/sigma;
                }
                break;

            default:
                assert(!"Incomplete swich statement. This is a bug. Please report.");
            }
        }
        return x;
    }

    double EdgeworthSeries1D::cdfFactor(const double xin) const
    {
        if (order_)
        {
            long double coeffs[MAXCOEFFS];
            unsigned maxdeg;
            if (slrMode_)
                hermiteCoeffsSLR(coeffs, &maxdeg);
            else
                hermiteCoeffsNormal(coeffs, &maxdeg);
            return hermiteSeriesSumProb(coeffs, maxdeg, normalizedCoord(xin));
        }
        else
            return 0.0;
    }

    double EdgeworthSeries1D::empiricalCentralMoment(const unsigned k) const
    {
        switch (k)
        {
        case 0U:
            return 1.0;

        case 1U:
            return 0.0;

        default:
            if (k > 200U)
                throw std::invalid_argument("In npstat::EdgeworthSeries1D::empiricalCentralMoment: "
                                            "moment order is too high");
            GaussHermiteQuadrature quad(128);
            EdgeworthSeriesMomentFcn fcn(*this, k);
            return quad.integrateProb(edgMean(), edgStdev(), fcn);
        }
    }

    void EdgeworthSeries1D::hermiteCoeffsSLR(long double* coeffs, unsigned* maxdeg) const
    {
        for (unsigned i=0; i<MAXCOEFFS; ++i)
            coeffs[i] = 0.0L;
        *maxdeg = 0U;

        assert(order_);
        const unsigned nk = cumulants_.size();
        assert(nk >= order_);
        *maxdeg = order_ - 1U;
        assert(MAXCOEFFS > *maxdeg);
        const double k1 = m_ == EDGEWORTH_SEVERINI ? cumulants_[0] : 0.0;
        coeffs[0] = k1;

        if (order_ > 1U)
        {
            const double k2 = m_ == EDGEWORTH_SEVERINI ? cumulants_[1] : 1.0;
            const double k2m1 = k2 - 1.0;
            coeffs[1] = (k1*k1 + k2m1)/2.0;
            
            if (order_ > 2U)
            {
                const double k3 = cumulants_[2];
                coeffs[2] = (k1*(k1*k1 + 3*k2m1) + k3)/6;
                
                if (order_ > 3U)
                {
                    const double k4 = cumulants_[3];
                    const double k1sq = k1*k1;
                    coeffs[3] = (k1sq*k1sq + 6*k1sq*k2m1 + 3*k2m1*k2m1 + 4*k1*k3 + k4)/24;
                    
                    if (order_ > 4U)
                        // We should not be here
                        assert(0);
                }

                if (m_ == EDGEWORTH_CLASSICAL)
                {
                    const double sigma = sqrt(cumulants_[1]);
                    for (unsigned i=2; i<=*maxdeg; ++i)
                        coeffs[i] /= powl(sigma, i);
                }
            }
        }
    }

    void EdgeworthSeries1D::hermiteCoeffsNormal(long double* coeffs, unsigned* maxdeg) const
    {
        for (unsigned i=0; i<MAXCOEFFS; ++i)
            coeffs[i] = 0.0L;
        *maxdeg = 0U;

        assert(order_);
        const unsigned nk = cumulants_.size();
        assert(nk > 2U);

        const double k1 = m_ == EDGEWORTH_SEVERINI ? cumulants_[0] : 0.0;
        const double k2 = m_ == EDGEWORTH_SEVERINI ? cumulants_[1] : 1.0;
        const double k3 = cumulants_[2];

        *maxdeg = 2U;
        assert(MAXCOEFFS > *maxdeg);
        coeffs[0] = k1;
        coeffs[2] = k3/6.0;

        if (order_ > 1U)
        {
            assert(nk > 3U);
            const double k4 = cumulants_[3];
            const double k2m1 = k2 - 1.0;

            *maxdeg = 5;
            assert(MAXCOEFFS > *maxdeg);
            coeffs[1] += (k1*k1 + k2m1)/2.0;
            coeffs[3] += (4*k1*k3 + k4)/24.0;
            coeffs[5] += k3*k3/72.0;

            if (order_ > 2U)
            {
                assert(nk > 4U);
                const double k5 = cumulants_[4];

                *maxdeg = 8;
                assert(MAXCOEFFS > *maxdeg);
                coeffs[2] += k1*(k1*k1 + 3*k2m1)/6;
                coeffs[4] += (10*k1*k1*k3 + 10*k2m1*k3 + 5*k1*k4 + k5)/120;
                coeffs[6] += (k3*(2*k1*k3 + k4))/144;
                coeffs[8] += k3*k3*k3/1296;

                if (order_ > 3U)
                {
                    assert(nk > 5U);
                    const double k6 = cumulants_[5];
                    const double k1sq = k1*k1;
                    const double k3sq = k3*k3;

                    *maxdeg = 11;
                    assert(MAXCOEFFS > *maxdeg);
                    coeffs[3] += (k1sq*k1sq + 6*k1sq*k2m1 + 3*k2m1*k2m1)/24;
                    coeffs[5] += (20*k1*k1sq*k3 + 60*k1*k2m1*k3 + 15*k1sq*k4 + 15*k2m1*k4 + 6*k1*k5 + k6)/720;
                    coeffs[7] += (40*k1sq*k3sq + 40*k2m1*k3sq + 40*k1*k3*k4 + 5*k4*k4 + 8*k3*k5)/5760;
                    coeffs[9] += (k3sq*(4*k1*k3 + 3*k4))/5184;
                    coeffs[11] += k3sq*k3sq/31104;

                    if (order_ > 4U)
                        assert(0);
                }
            }

            if (m_ == EDGEWORTH_CLASSICAL)
            {
                const double sigma = sqrt(cumulants_[1]);
                for (unsigned i=2; i<=*maxdeg; ++i)
                    coeffs[i] /= powl(sigma, i);
            }
        }
    }

    double EdgeworthSeriesMomentFcn::operator()(const double& x) const
    {
        const double fact = series_.densityFactor(x);
        switch (degree_)
        {
        case 0:
            return fact;

        case 1:
            return x*fact;

        default:
            return pow(x - mean_, degree_)*fact;
        }
    }
}
